
#include<stdio.h>
#include<assert.h>
#include<string.h>

#include"nb_key.h"

#include<HD/hd.h>
#include<HDU/hduError.h>
#include<HDU/hduVector.h>

/*
 * Program tworzy sfere, w ktorej wystepuje tarcie (Kulombowskie).
 */

HDCallbackCode HDCALLBACK frictionCallback(void *data);

int main(int argc, char* argv[])
{    
    HDErrorInfo error;
    HDSchedulerHandle hFriction;

    /* Inicjalizacja urzadzenia. Od tego zaczynamy */

    HHD hHD = hdInitDevice(HD_DEFAULT_DEVICE);
    if (HD_DEVICE_ERROR(error = hdGetError())) {
        hduPrintError(stderr, &error, "Failed to initialize haptic device");
        return -1;
    }

	/* Udalo sie. Urzadzenie jest i dziala. */
    printf("Hello Haptic Device!\n");
    printf("Found device model: %s.\n\n", hdGetString(HD_DEVICE_MODEL_TYPE));

	/* Start petli urzadzenia */
    hdEnable(HD_FORCE_OUTPUT);
    hdStartScheduler();

    /* Czy sie udalo? */
    if (HD_DEVICE_ERROR(error = hdGetError())) {
        hduPrintError(stderr, &error, "Failed to start scheduler");
        fprintf(stderr, "\nPress any key to quit.\n");
        return -1;
    }

	/* Kalibracja urzadzenia. Co bedzie, jesli sie ja wylaczy? */
	calibrate_haptic();

    /* Instalujemy callback. 0 oznacza pusty wskaznik do danych, ktore moglyby byc zwrocone. */
    hFriction = hdScheduleAsynchronous(frictionCallback, 0, HD_MAX_SCHEDULER_PRIORITY);
    printf("Feel around for the friction bubble...\n");
    printf("Press any key to quit.\n\n");

    int done=0;
 
    /* Przestawiamy terminal w tryb nieblokujacy.  */
    nonblock(NB_ENABLE);
    while(!done) {
        usleep(100000);
        int i=kbhit();
		if(i!=0) { /* Uzytkownik nacisnal klawisz. Koniec. */
			break;
		}
        /* Sprawdzamy, czy callback sie nie zakonczyl */
        if (!hdWaitForCompletion(hFriction, HD_WAIT_CHECK_STATUS)) {
            fprintf(stderr, "Callback terminated.\n");     
            break;
        }
    }
    nonblock(NB_DISABLE);


    /* Zatrzymanie schedulera i usuniecie callback'u */
    hdStopScheduler();
    hdUnschedule(hFriction);

    /* Wylaczenie urzadzenia. */
    hdDisableDevice(hHD);

    return 0;
}

/*
 * Callback, ktory jest wywolywany w kazdej iteracji petli schedulera.
 */
HDCallbackCode HDCALLBACK frictionCallback(void *data)
{
    const HDdouble f=.5; /* N */
	const HDdouble radius=25; /* mm */

    HDErrorInfo error;
    hduVector3Dd position;
    hduVector3Dd last_position;

    HHD hHD = hdGetCurrentDevice();

	/* Poczatek ramki haptycznej */
    hdBeginFrame(hHD);

    /* Biezaca pozycja urzadzenia */
    hdGetDoublev(HD_CURRENT_POSITION, position);
    hdGetDoublev(HD_LAST_POSITION, last_position);
    
	/* Odleglosc od srodka sfery (0, 0, 0). */
	double modX=hduVecMagnitude(position);

    hduVector3Dd force;

	memset(force, 0, sizeof(force));

	if(modX<=radius ) {
		hduVector3Dd velocity;

		/* Przyblizamy predkosc, jako roznice pomiedzy kolejnymi polozeniami. */
		hduVecSubtract(velocity, position, last_position);

		double modV=hduVecMagnitude(velocity);

		if(modV>0) { // Warunek, zeby uniknac dzielenia przez 0
			hduVecScale(force, velocity, -f/modV);
		}
	}

    /* Aktualizacja sily. */
    hdSetDoublev(HD_CURRENT_FORCE, force);
    
    /* Koniec ramki. */
    hdEndFrame(hHD);

    /* Sprawdzenie, czy nie wystapil blad. */
    if (HD_DEVICE_ERROR(error = hdGetError())) {
        hduPrintError(stderr, &error, "Error detected while rendering frictionless surface\n");
        
        if (hduIsSchedulerError(&error)) {
            return HD_CALLBACK_DONE;
        }
    }

    /* Ten callback ma byc wywolany rowniez w nastepnej iteracji schedulera. */
    return HD_CALLBACK_CONTINUE;
}

