
#include<stdio.h>
#include<assert.h>
#include<string.h>

#include"nb_key.h"
#include"calibration.h"

#include<HD/hd.h>
#include<HDU/hduError.h>
#include<HDU/hduVector.h>

/*
 * Program tworzy studnie grawitacyjna (prawdziwa ;-)
 */

HDCallbackCode HDCALLBACK gravityWellCallback(void *data);

int main(int argc, char* argv[])
{    
    HDErrorInfo error;
    HDSchedulerHandle hGravityWell;

    /* Inicjalizacja urzadzenia. Od tego zaczynamy */

    HHD hHD = hdInitDevice(HD_DEFAULT_DEVICE);
    if (HD_DEVICE_ERROR(error = hdGetError())) {
        hduPrintError(stderr, &error, "Failed to initialize haptic device");
        return -1;
    }

	/* Udalo sie. Urzadzenie jest i dziala. */
    printf("Hello Haptic Device!\n");
    printf("Found device model: %s.\n\n", hdGetString(HD_DEVICE_MODEL_TYPE));


	/* Start petli urzadzenia */
    hdEnable(HD_FORCE_OUTPUT);
    hdStartScheduler();

    /* Czy sie udalo? */
    if (HD_DEVICE_ERROR(error = hdGetError())) {
        hduPrintError(stderr, &error, "Failed to start scheduler");
        fprintf(stderr, "\nPress any key to quit.\n");
        return -1;
    }

	/* Kalibracja urzadzenia. Co bedzie, jesli sie ja wylaczy? */
	//calibrate_haptic();

    HDdouble maxWorkspace[6];
    hdGetDoublev(HD_MAX_WORKSPACE_DIMENSIONS, maxWorkspace);

	printf("%f %f %f\n", maxWorkspace[0], maxWorkspace[1], maxWorkspace[2]);
	printf("%f %f %f\n", maxWorkspace[3], maxWorkspace[4], maxWorkspace[5]);



    /* Instalujemy callback. 0 oznacza pusty wskaznik do danych, ktore moglyby byc zwrocone. */
    hGravityWell = hdScheduleAsynchronous(gravityWellCallback, 0, HD_MAX_SCHEDULER_PRIORITY);
    printf("Feel around for the gravity well...\n");
    printf("Press any key to quit.\n\n");

    int done=0;
 
    /* Przestawiamy terminal w tryb nieblokujacy.  */
    nonblock(NB_ENABLE);
    while(!done) {
        usleep(100000);
        int i=kbhit();
		if(i!=0) { /* Uzytkownik nacisnal klawisz. Koniec. */
			break;
		}
        /* Sprawdzamy, czy callback sie nie zakonczyl */
        if (!hdWaitForCompletion(hGravityWell, HD_WAIT_CHECK_STATUS)) {
            fprintf(stderr, "Callback terminated.\n");     
            break;
        }
    }
    nonblock(NB_DISABLE);


    /* Zatrzymanie schedulera i usuniecie callback'u */
    hdStopScheduler();
    hdUnschedule(hGravityWell);

    /* Wylaczenie urzadzenia. */
    hdDisableDevice(hHD);

    return 0;
}


/*
 * Callback, ktory jest wywolywany w kazdej iteracji petli schedulera.
 */
HDCallbackCode HDCALLBACK gravityWellCallback(void *data)
{
    const HDdouble G= 100; /* N/mm */
    const HDdouble kGravityWellInfluence = 100; /* mm */
    const HDdouble kGravityWellSize = 10; /* mm */
    const HDdouble kGravityWellSoftSize = 11; /* mm */

    /* Polozenie studni. */
    static const hduVector3Dd wellPos = {0,0,0};

    HDErrorInfo error;
    hduVector3Dd position;
    hduVector3Dd force;
    hduVector3Dd positionTwell;

    HHD hHD = hdGetCurrentDevice();

	/* Poczatek ramki haptycznej */
    hdBeginFrame(hHD);

    /* Biezaca pozycja urzadzenia */
    hdGetDoublev(HD_CURRENT_POSITION, position);
    
    memset(force, 0, sizeof(hduVector3Dd));
    
    /* >  positionTwell = wellPos-position  < 
       Wektor od biezacego polozenia do srodka studni */
    hduVecSubtract(positionTwell, wellPos, position);
    
	/* Obliczanie sily   */
    if (hduVecMagnitude(positionTwell) < kGravityWellInfluence && kGravityWellSize< hduVecMagnitude(positionTwell)) {
        /* >  F = x * G / |x|^3  < 
           F: Sila (N)
           G: "stala grawitacyjna" (N/mm^2)
           x: Wektor od polozenia urzadzenia do studni */

		double modX=hduVecMagnitude(positionTwell);
		double scale=G/(modX*modX*modX);

		/* "Zmiekczamy" scianke sfery wewnatrz, ktorej nie ma grawitacji */

		if(modX<kGravityWellSoftSize) scale*=(modX-kGravityWellSize)/(kGravityWellSoftSize-kGravityWellSize);

        hduVecScale(force, positionTwell, scale);
    }

    /* Aktualizacja sily. */
    hdSetDoublev(HD_CURRENT_FORCE, force);
    
    /* Koniec ramki. */
    hdEndFrame(hHD);

    /* Sprawdzenie, czy nie wystapil blad. */
    if (HD_DEVICE_ERROR(error = hdGetError())) {
        hduPrintError(stderr, &error, "Error detected while rendering gravity well\n");
        
        if (hduIsSchedulerError(&error)) {
            return HD_CALLBACK_DONE;
        }
    }

    /* Ten callback ma byc wywolany rowniez w nastepnej iteracji schedulera. */
    return HD_CALLBACK_CONTINUE;
}
